import { __extends } from 'tslib';
import { ChangeDetectorRef } from '@angular/core';
import { Subject } from 'rxjs';
import { filter, startWith, map } from 'rxjs/operators';
import { SubscriptionManager } from 's-rxjs-utils';
import { NG_VALUE_ACCESSOR, FormControl } from '@angular/forms';
import { noop } from 'micro-dash';

/**
 * Use as the superclass for anything managed by angular's dependency injection for care-free use of `subscribeTo()`. It simply calls `unsubscribe()` during `ngOnDestroy()`. If you override `ngOnDestroy()` in your subclass, be sure to invoke the super implementation.
 *
 * ```ts
 * @Injectable()
 * // or @Component() (also consider DirectiveSuperclass)
 * // or @Directive() (also consider DirectiveSuperclass)
 * // or @Pipe()
 * class MyThing extends InjectableSuperclass {
 *   constructor(somethingObservable: Observable) {
 *     super();
 *     this.subscribeTo(somethingObservable);
 *   }
 *
 *   ngOnDestroy() {
 *     // if you override ngOnDestroy, be sure to call this too
 *     super.ngOnDestroy();
 *   }
 * }
 * ```
 */
var InjectableSuperclass = /** @class */ (function (_super) {
    __extends(InjectableSuperclass, _super);
    function InjectableSuperclass() {
        var _this = _super.call(this) || this;
        _this.destructionSubject = new Subject();
        _this.destruction$ = _this.destructionSubject.asObservable();
        return _this;
    }
    InjectableSuperclass.prototype.ngOnDestroy = function () {
        this.unsubscribe();
        this.destructionSubject.next();
        this.destructionSubject.complete();
    };
    return InjectableSuperclass;
}(SubscriptionManager));

/**
 * Extend this when creating a directive (including a component, which is a kind of directive) to gain access to the helpers demonstrated below. **Warning:** You _must_ include a constructor in your subclass.
 *
 * ```ts
 * @Component({
 *   selector: "s-color-text",
 *   template: `
 *     <span [style.background]="color">{{ color }}</span>
 *   `,
 *   // note that `bindToInstance()` works even with OnPush change detection
 *   changeDetection: ChangeDetectionStrategy.OnPush,
 * })
 *  class ColorTextComponent extends DirectiveSuperclass {
 *   @Input() prefix?: string;
 *   @Input() prefix2?: string;
 *   color!: string;
 *
 *   // Even if you don't need extra arguments injector, you must still include a constructor. It is required for angular to provide `Injector`.
 *   constructor(
 *     @Inject("color$") color$: Observable<string>,
 *     injector: Injector,
 *   ) {
 *     super(injector);
 *
 *     // combine everything to calculate `color` and keep it up to date
 *     this.bindToInstance(
 *       "color",
 *       combineLatest(
 *         this.getInput$("prefix"),
 *         this.getInput$("prefix2"),
 *         color$,
 *       ).pipe(map((parts) => parts.filter((p) => p).join(""))),
 *     );
 *   }
 * }
 * ```
 */
var DirectiveSuperclass = /** @class */ (function (_super) {
    __extends(DirectiveSuperclass, _super);
    function DirectiveSuperclass(injector) {
        var _this = _super.call(this) || this;
        /**
         *  Emits the set of `@Input()` property names that change during each call to `ngOnChanges()`.
         */
        _this.inputChanges$ = new Subject();
        _this.changeDetectorRef = injector.get(ChangeDetectorRef);
        return _this;
    }
    DirectiveSuperclass.prototype.ngOnChanges = function (changes) {
        this.inputChanges$.next(new Set(Object.getOwnPropertyNames(changes)));
    };
    /**
     * @return an observable of the values for one of this directive's `@Input()` properties
     */
    DirectiveSuperclass.prototype.getInput$ = function (key) {
        var _this = this;
        return this.inputChanges$.pipe(filter(function (keys) { return keys.has(key); }), startWith(undefined), map(function () { return _this[key]; }));
    };
    /**
     * Binds an observable to one of this directive's instance variables. When the observable emits the instance variable will be updated, and change detection will be triggered to propagate any changes. Use this an an alternative to repeating `| async` multiple times in your template.
     */
    DirectiveSuperclass.prototype.bindToInstance = function (key, value$) {
        var _this = this;
        this.subscribeTo(value$, function (value) {
            _this[key] = value;
            _this.changeDetectorRef.markForCheck();
        });
    };
    return DirectiveSuperclass;
}(InjectableSuperclass));

/**
 * Use in the `providers` of a component that implements `ControlValueAccessor` to reduce some boilerplate.
 *
 * ```ts
 * @Component({ providers: [provideValueAccessor(MyFormControl)] }
 * class MyFormControl extends BaseFormControl {
 *   // ...
 * }
 * ```
 */
function provideValueAccessor(type) {
    return {
        provide: NG_VALUE_ACCESSOR,
        useExisting: type,
        multi: true,
    };
}
/**
 * Extend this when creating a form control to reduce some boilerplate. **Warning:** You _must_ include a constructor in your subclass.
 *
 * This example allows 2-way binding to a number via `[(ngModel)]`, `[formControl]`, or any other technique that leverages the `ControlValueAccessor` interface.
 * ```ts
 * @Component({
 *   template: `
 *     <button (click)="increment()" [disabled]="isDisabled">{{ counter }}</button>
 *   `,
 *   providers: [provideValueAccessor(CounterComponent)],
 * })
 * class CounterComponent extends FormControlSuperclass<number> {
 *   counter = 0;
 *
 *   // This looks unnecessary, but is required for Angular to provide `Injector`
 *   constructor(injector: Injector) {
 *     super(injector);
 *   }
 *
 *   handleIncomingValue(value: number) {
 *     this.counter = value;
 *   }
 *
 *   increment() {
 *     this.emitOutgoingValue(++this.counter);
 *     this.onTouched();
 *   }
 * }
 * ```
 */
var FormControlSuperclass = /** @class */ (function (_super) {
    __extends(FormControlSuperclass, _super);
    function FormControlSuperclass() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        /** Call this to emit a new value when it changes. */
        _this.emitOutgoingValue = noop;
        /** Call this to "commit" a change, traditionally done e.g. on blur. */
        _this.onTouched = noop;
        /** You can bind to this in your template as needed. */
        _this.isDisabled = false;
        return _this;
    }
    /** Called as angular propagates value changes to this `ControlValueAccessor`. You normally do not need to use it. */
    FormControlSuperclass.prototype.writeValue = function (value) {
        this.handleIncomingValue(value);
        this.changeDetectorRef.markForCheck();
    };
    /** Called as angular sets up the binding to this `ControlValueAccessor`. You normally do not need to use it. */
    FormControlSuperclass.prototype.registerOnChange = function (fn) {
        this.emitOutgoingValue = fn;
    };
    /** Called as angular sets up the binding to this `ControlValueAccessor`. You normally do not need to use it. */
    FormControlSuperclass.prototype.registerOnTouched = function (fn) {
        this.onTouched = fn;
    };
    /** Called as angular propagates disabled changes to this `ControlValueAccessor`. You normally do not need to use it. */
    FormControlSuperclass.prototype.setDisabledState = function (isDisabled) {
        this.isDisabled = isDisabled;
        this.changeDetectorRef.markForCheck();
    };
    return FormControlSuperclass;
}(DirectiveSuperclass));

/**
 * Extend this when creating a form control that simply wraps an existing form control, to reduce a lot of boilerplate. **Warning:** You _must_ include a constructor in your subclass.
 *
 * Example when you don't need to modify the wrapped control's value:
 * ```ts
 * @Component({
 *   template: `<input [formControl]="formControl">`,
 *   providers: [provideValueAccessor(StringComponent)],
 * })
 * class StringComponent extends WrappedFormControlSuperclass<string> {
 *   // This looks unnecessary, but is required for Angular to provide `Injector`
 *   constructor(injector: Injector) {
 *     super(injector);
 *   }
 * }
 * ```
 *
 * Example when you need to modify the wrapped control's value:
 * ```ts
 * @Component({
 *   template: `<input type="datetime-local" [formControl]="formControl">`,
 *   providers: [provideValueAccessor(DateComponent)],
 * })
 * class DateComponent extends WrappedFormControlSuperclass<Date, string> {
 *   // This looks unnecessary, but is required for Angular to provide `Injector`
 *   constructor(injector: Injector) {
 *     super(injector);
 *   }
 *
 *   protected innerToOuter(value: string): Date {
 *     return new Date(value + "Z");
 *   }
 *
 *   protected outerToInner(value: Date): string {
 *     if (value === null) {
 *       return ""; // happens during initialization
 *     }
 *     return value.toISOString().substr(0, 16);
 *   }
 * }
 * ```
 */
var WrappedFormControlSuperclass = /** @class */ (function (_super) {
    __extends(WrappedFormControlSuperclass, _super);
    function WrappedFormControlSuperclass(injector) {
        var _this = _super.call(this, injector) || this;
        /** Bind this to your inner form control to make all the magic happen. */
        _this.formControl = new FormControl();
        _this.subscribeTo(_this.formControl.valueChanges, function (value) {
            _this.emitOutgoingValue(_this.innerToOuter(value));
        });
        _this.formControl.markAsTouched = function () {
            _this.onTouched();
        };
        return _this;
    }
    /** Called as angular propagates values changes to this `ControlValueAccessor`. You normally do not need to use it. */
    WrappedFormControlSuperclass.prototype.handleIncomingValue = function (value) {
        this.formControl.setValue(this.outerToInner(value), { emitEvent: false });
    };
    /** Called as angular propagates disabled changes to this `ControlValueAccessor`. You normally do not need to use it. */
    WrappedFormControlSuperclass.prototype.setDisabledState = function (isDisabled) {
        if (isDisabled) {
            this.formControl.disable({ emitEvent: false });
        }
        else {
            this.formControl.enable({ emitEvent: false });
        }
        _super.prototype.setDisabledState.call(this, this.isDisabled);
    };
    /** Override this to modify a value coming from the outside to the format needed within this component. */
    WrappedFormControlSuperclass.prototype.outerToInner = function (value) {
        return value;
    };
    /** Override this to modify a value coming from within this component to the format expected on the outside. */
    WrappedFormControlSuperclass.prototype.innerToOuter = function (value) {
        return value;
    };
    return WrappedFormControlSuperclass;
}(FormControlSuperclass));

/*
 * Public API Surface of s-ng-utils
 */

/**
 * Generated bundle index. Do not edit.
 */

export { DirectiveSuperclass, FormControlSuperclass, InjectableSuperclass, WrappedFormControlSuperclass, provideValueAccessor };
//# sourceMappingURL=s-ng-utils.js.map
